<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.

/**
 * This file contains the data dump generator for the AAT plugin.
 *
 * @package     local_aatplugin
 * @copyright   2023 Joseph Larochelle
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
namespace local_aatplugin;
use Exception;
use RuntimeException;
/**
 * Class to generate the SQL dump for the AAT plugin.
 */
class data_dump {
    /**
     * The moodle database object.
     *
     * @var moodle_database
     */
    private $db;
    /**
     * The moodle configuration object.
     *
     * @var stdClass
     */
    private $cfg;
    /**
     * The relations mapper object.
     *
     * @var coursetable_relations_mapper
     */
    private coursetable_relations_mapper $relationsmapper;
    /**
     * The database schema generator object.
     *
     * @var database_schema_generator
     */
    private database_schema_generator $databaseschemagenerator;
    /**
     * Constructor for the data dump class.
     *
     * @param moodle_database $DB The moodle database object.
     * @param stdClass $CFG The moodle configuration object.
     */
    public function __construct($DB, $CFG) {
        $this->db = $DB;
        $this->cfg = $CFG;
        $this->relationsmapper = new coursetable_relations_mapper();
        $this->databaseschemagenerator = new database_schema_generator();
    }
    /**
     * Generate an SQL dump for the given courseids and write it to the specified file.
     *
     * @param array $courseids An array of course ids.
     * @param string $filepath The path to the file to write the SQL dump to.
     */
    public function generate_sql_dump(array $courseids, string $filepath) {
        // Extract all the keys from the array of courseids.
        $courseids = array_values($courseids);
        // Initialize the sql dump.
        // Add a comment to the sql dump file with the moodle version.
        $sqldump = "-- AAT plugin SQL dump\n";
        $sqldump .= "-- {$this->cfg->release}\n\n";

        // Check if the combined_XML_files.xml file exists.
        if (!file_exists($this->cfg->tempdir . '/aatplugin/combined_XML_files.xml')) {
            // Create the file and folder.
            if (!mkdir($concurrentdirectory = $this->cfg->tempdir . '/aatplugin') && !is_dir($concurrentdirectory)) {
                throw new RuntimeException(sprintf(get_string('directory_not_created', 'local_aatplugin'), $concurrentdirectory));
            }
            // If it does not exist, create it.
            local_aatplugin_combinexmlfiles($this->cfg->dirroot);
        }
        // Check if the schema file exists.
        if (!file_exists($this->cfg->tempdir . '/aatplugin/database_schema.sql')) {
            // If it does not exist, create it.
            // Load and parse the XML file.
            $xmlfilepath = $this->cfg->tempdir . '/aatplugin/combined_XML_files.xml';
            $xml = simplexml_load_file($xmlfilepath);
            // Open a file to write the SQL statements.
            $sqlfile = fopen($this->cfg->tempdir ."/aatplugin/database_schema.sql", 'wb');
            // Generate SQL for each table and write to the file.
            foreach ($xml->TABLES->TABLE as $table) {
                $sqlstatement = $this->databaseschemagenerator->generatecreatetablesql($table);
                fwrite($sqlfile, $sqlstatement);
            }
            // Close the file.
            fclose($sqlfile);
        }
        // Add the schema file to the sql dump.
        $schema = file_get_contents($this->cfg->tempdir . '/aatplugin/database_schema.sql');
        $sqldump .= $schema;
        // Check to see if the related_tables.txt file exists.
        if (!file_exists($this->cfg->tempdir .'/aatplugin/related_tables.txt')) {
            // If it does not exist, create it.
            // Path to your XML file.
            $xmlfilepath = $this->cfg->tempdir . '/aatplugin/combined_XML_files.xml';
            $xml = simplexml_load_file($xmlfilepath);

            $initialtables = $this->relationsmapper->findinitialtables($xml);
            $relatedtableswithparent = $this->relationsmapper->findallrelatedtables($xml, $initialtables);

            // Print the related tables with their direct parent.
            $serializedoutput = serialize($relatedtableswithparent);
            file_put_contents($this->cfg->tempdir .'/aatplugin/related_tables.txt', $serializedoutput);
        }
        $serializedtables = file_get_contents($this->cfg->tempdir .'/aatplugin/related_tables.txt');
        $relatedtables = unserialize($serializedtables, ["allowed_classes" => false]);

        // Iterate through the array and collect all tables that do not have a subtable.
        // These tables are direct children of the course table.
        $coursetables = [];
        foreach ($relatedtables as $table => $subtable) {
            // If the table has no subtable, add it to the array of course tables.
            if ($subtable == null) {
                $coursetables[$table] = [];
                // Remove it from the array of related tables.
                unset($relatedtables[$table]);
            }
        }
        unset($coursetables['question']);
        // All the tables in $coursetables are direct children of the course table.
        // We iterate through the entire array and add all the records that are related to the courses.
        // Listed in the $courseids array.
        foreach ($coursetables as $table => $ids) {
            // Prepend the 'mdl_' prefix to the table name.
            $prefixedtablename = $this->cfg->prefix . $table;
            // Get the columns of the table.
            try {
                $columns = $this->db->get_columns($table);
                $field = array_key_exists('courseid', $columns) ? 'courseid' : 'course';
                // Fetch records related to the course.
                $records = $this->get_table_records($table, $field, $courseids);
                $sqldump .= $this->create_sql_statements($records, $table, $coursetables);
            } catch (Exception $e) {
                echo sprintf(get_string('caught_exception', 'local_aatplugin'), $e->getMessage()) . "\n";
                echo '<br>';
            }
        }

        // Fetch all the rows in the mdl_course table matching the courseids.
        $coursesrecords = $this->get_table_records('course', 'id', $courseids);
        // Add those records to the sql dump.
        $sqldump .= $this->create_sql_statements($coursesrecords, 'course', $coursetables);

        // Add all rows from the mdl_config table to the sql dump.
        $configsrecords = $this->db->get_records('config');
        $sqldump .= $this->create_sql_statements($configsrecords, 'config', $coursetables);

        // Get all the distinct userids from the logstore_standard_log table where courseid is in the array of courseids.
        $logstorestandardlogrecords = $this->db->get_records_list('logstore_standard_log', 'courseid', $courseids);
        $userids = [];
        foreach ($logstorestandardlogrecords as $logstorestandardlogrecord) {
            if (!empty($logstorestandardlogrecord->userid)) {
                $userids[$logstorestandardlogrecord->userid] = true;
            }
            if (!empty($logstorestandardlogrecord->relateduserid)) {
                $userids[$logstorestandardlogrecord->relateduserid] = true;
            }
        }
        // Get all the records from the user table where the userid is in the array of userids.
        $usersrecords = $this->db->get_records_list('user', 'id', array_keys($userids));
        $sqldump .= $this->create_sql_statements($usersrecords, 'user', $coursetables);
        // Get all the records from the question table where createdby or modifiedby is in the array of userids.
        $questionsrecords = $this->db->get_records_list('question', 'createdby', array_keys($userids));
        $sqldump .= $this->create_sql_statements($questionsrecords, 'question', $coursetables);
        $questionsrecords = $this->db->get_records_list('question', 'modifiedby', array_keys($userids));
        $sqldump .= $this->create_sql_statements($questionsrecords, 'question', $coursetables);
        // Get all the records from the context table.
        $contextsrecords = $this->db->get_records('context');
        $sqldump .= $this->create_sql_statements($contextsrecords, 'context', $coursetables);

        // While there are tables in the array of tables to be added to the sql dump.
        for ($i = 0; $i < 3; $i++) {
            $errors = 0;
            // Iterate through the array of table in relatedtables.
            foreach ($relatedtables as $table => $subtable) {
                // Check if the key of $subtable is in the array of keys of $coursetables.
                if (array_key_exists(key($subtable), $coursetables)) {
                    // Prepend the 'mdl_' prefix to the table name.
                    $prefixedtablename = $this->cfg->prefix . $table;
                    // Get the value of the key of $subtable.
                    $field = $subtable[key($subtable)];
                    try {
                        // Fetch records related to the course.
                        $records = $this->get_table_records($table, $field, $coursetables[key($subtable)]);
                    } catch (Exception $e) {
                        // If there is an exception, print the table name. bbbtext_simple is ignored.
                        if ($table == 'bbbext_simple')
                            continue;
                        if ($errors == 0) {
                            echo get_string('dumpfileerror', 'local_aatplugin') . '<br>';
                            $errors = 1;
                        }
                        // Print the table name.
                        echo '<br>';
                        echo $table;
                        echo '<br>';
                    }
                    $sqldump .= $this->create_sql_statements($records, $table, $coursetables);
                    // Remove the table from the array of related tables.
                    unset($relatedtables[$table]);
                }
            }
        }
        // This section is used to add all the tables with data that is not directly related to courses but is used in AAT.
        // Because of the lack of relationship they need to be hardcoded into the plugin.
        // Book chapters.
        $bookchaptersrecords = $this->db->get_records_list('book_chapters', 'bookid', $coursetables['book']);
        $sqldump .= $this->create_sql_statements($bookchaptersrecords, 'course', $coursetables);
        // Return all the records from the course_categories table.
        $coursecategoriesrecords = $this->db->get_records('course_categories');
        $sqldump .= $this->create_sql_statements($coursecategoriesrecords, 'course_categories', $coursetables);
        // Return the records from the course_modules_completion table.
        $coursemodulescompletionrecords = $this->db->get_records_list
        ('course_modules_completion', 'coursemoduleid', $coursetables['course_modules']);
        $sqldump .= $this->create_sql_statements($coursemodulescompletionrecords, 'course_modules_completion', $coursetables);
        // Return the records from the files table.
        $filesrecords = $this->db->get_records_list('files', 'userid', $coursetables['user']);
        $sqldump .= $this->create_sql_statements($filesrecords, 'files', $coursetables);
        // Add the records from repository_instances table.
        $repositoryinstancesrecords = $this->db->get_records_list('repository_instances', 'userid', $coursetables['user']);
        $sqldump .= $this->create_sql_statements($repositoryinstancesrecords, 'repository_instances', $coursetables);
        // Add the records from file_references table.
        if (!empty($coursetables['repository_instances'])) {
            $filereferencesrecords = $this->db->get_records_list('file_references', 'repositoryid',
                    $coursetables['repository_instances']);
            $sqldump .= $this->create_sql_statements($filereferencesrecords, 'file_references', $coursetables);
        }
        // Add the records from forum_read table.
        $forumreadrecords = $this->db->get_records_list('forum_read', 'forumid', $coursetables['forum']);
        $sqldump .= $this->create_sql_statements($forumreadrecords, 'forum_read', $coursetables);
        // Add the records from forum_track_prefs table.
        $forumtrackprefsrecords = $this->db->get_records_list('forum_track_prefs', 'forumid', $coursetables['forum']);
        $sqldump .= $this->create_sql_statements($forumtrackprefsrecords, 'forum_track_prefs', $coursetables);
        // Add the records from lti_submissions table.
        $ltisubmissionsrecords = $this->db->get_records_list('lti_submission', 'ltiid', $coursetables['lti']);
        $sqldump .= $this->create_sql_statements($ltisubmissionsrecords, 'lti_submission', $coursetables);
        // Add records from wiki_synonyms table.
        if (!empty($coursetables['wiki_subwikis'])) {
            $wikisynonymsrecords = $this->db->get_records_list('wiki_synonyms', 'subwikiid', $coursetables['wiki_subwikis']);
            $sqldump .= $this->create_sql_statements($wikisynonymsrecords, 'wiki_synonyms', $coursetables);
        }
        // Add all entries from the role table.
        $rolesrecords = $this->db->get_records('role');
        $sqldump .= $this->create_sql_statements($rolesrecords, 'role', $coursetables);
        // Get all the records from the role_assignments table where the userid is in the array of userids.
        $roleassignmentsrecords = $this->db->get_records_list('role_assignments', 'userid', array_keys($userids));
        $sqldump .= $this->create_sql_statements($roleassignmentsrecords, 'role_assignments', $coursetables);

        // Load the public key.
        $publickey = openssl_get_publickey(file_get_contents($this->cfg->dirroot . '/local/aatplugin/public_key.pem'));

        if (!$publickey) {
            throw new Exception('Failed to load public key');
        }

        // Define the key size.
        $keydetails = openssl_pkey_get_details($publickey);
        $keysize = $keydetails['bits'] / 8 - 11; // Key size in bytes.
        // Break the data into chunks and encrypt each one.
        $chunks = str_split($sqldump, $keysize);
        $encrypted = '';

        foreach ($chunks as $chunk) {
            $encryptedchunk = '';
            if (!openssl_public_encrypt($chunk, $encryptedchunk, $publickey)) {
                throw new Exception(get_string('encryptionerror', 'local_aatplugin'));
            }
            $encrypted .= $encryptedchunk;
        }

        // Write the encrypted data to the file.
        file_put_contents($filepath, $encrypted);
    }
    /**
     * Get records from a table where a field matches a value.
     *
     * @param string $table The name of the table.
     * @param string $field The name of the field.
     * @param array $values An array of values to match.
     * @return array An array of records.
     * @throws Exception If the records cannot be retrieved.
     */
    private function get_table_records($table, $field, $values) {
        try {
            return $records = $this->db->get_records_list($table, $field, $values);
        } catch (\dml_exception $e) {
            throw new \Exception(get_string('recordserror', 'local_aatplugin') . "{$table}", 0, $e);
        }
    }
    /**
     * Create the SQL statements for the records.
     *
     * @param array $records An array of records.
     * @param string $table The name of the table.
     * @param array $coursetables An array of course tables.
     * @return string The SQL statements.
     */
    private function create_sql_statements($records, $table, &$coursetables) {
        // Check to see if the records are empty.
        if (!empty($records)) {
            $prefixedtablename = $this->cfg->prefix . $table;
            $sqldump = "";
            foreach ($records as $record) {
                // Get the value of the id field.
                $id = $record->id;
                // Add the id to $courseTables.
                $coursetables[$table][] = $id;

                $columns = array_keys(get_object_vars($record));
                $valuestrings = [];

                foreach ($record as $columnname => $value) {
                    $escapedvalue = ($value === null) ? 'NULL' : "E'" . addslashes($value) . "'";
                    $valuestrings[] = $escapedvalue;
                }

                $sqldump .= "INSERT INTO " . $prefixedtablename . "(" . implode(", ", $columns) . ") VALUES (" .
                        implode(",", $valuestrings) .
                        ");\n ";
            }
            return $sqldump;
        }
    }
}
