<?php
// This file is part of Moodle - https://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <https://www.gnu.org/licenses/>.
// Include the main configuration file of the Moodle site.

/**
 * This file is responsible for downloading a file from the server.
 *
 * @package     local_aatplugin
 * @copyright   2023 Joseph Larochelle
 * @license     https://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
require_once(__DIR__ . '/../../config.php');
require_login();

// Define the directory where downloadable files are stored.
$download_dir = $CFG->tempdir . '/aatplugin/';

// Retrieve the file name from the request parameters.
$filename = required_param('file', PARAM_FILE);

// Sanitize the file name to prevent directory traversal attacks.
$sanitized_filename = basename($filename);

// Construct the full file path.
$filepath = $download_dir . $sanitized_filename;

// Check if the specified file exists and is within the download directory.
if (file_exists($filepath) && strpos(realpath($filepath), realpath($download_dir)) === 0) {
    // Set headers to initiate a file download.
    header('Content-Description: File Transfer');
    header('Content-Type: application/octet-stream'); // Specify the content type as binary/octet-stream for arbitrary binary data.
    header('Content-Disposition: attachment; filename="' . $sanitized_filename . '"'); // Suggest a filename for the downloaded file.
    header('Expires: 0'); // Indicate that the file should not be cached.
    header('Cache-Control: must-revalidate'); // HTTP 1.1. Cache-Control header for revalidation and reload.
    header('Pragma: public'); // Backward compatibility with HTTP 1.0 for public cache.
    header('Content-Length: ' . filesize($filepath)); // Specify the size of the file being downloaded.

    ob_end_clean(); // Clean (erase) the output buffer and turn off output buffering.
    // This is necessary to avoid any additional output that could corrupt the file stream.

    readfile($filepath); // Output the file content for download.

    // Delete the file after download to prevent accumulation on the server.
    unlink($filepath);

    exit;
} else {
    // If the file does not exist, display an error message.
    echo $filepath;
    echo get_string('downloaderror', 'local_aatplugin');
}
